#include "GraphicsGems.h"
#include <stdlib.h>  /* For debugging only -- CAS */
#include <stdio.h>
#include "geosim.h"  /* For debugging only -- CAS */

/********************************************************************
Compute the transformation matrix from triangle P
to triangle Q. Return TRUE if the transformation is possible,
FALSE if triangle P is degenerate, or segment (Q1, Q2) is degenerate.
Entry:
  Q1, Q2, Q3: vertices for triangle Q.
  P1, P2, P3: vertices for triangle P.
Exit:
  Trans: resulting transformation matrix.
*********************************************************************/

boolean triangle_transform(Point2 P1, Point2 P2, Point2 P3,
                           Point2 Q1, Point2 Q2, Point2 Q3, double Trans[3][3])
{
  Point2 P2P, P3P, Q2P, Q3P;
  double LP2Psqrd, LQ2Psqrd, A, B, C;
  double tempx, tempy, temp;

  /* First, deal with degeneracies that we can handle */
  if ((Q1.x == Q2.x) && (Q2.x == Q3.x) &&
      (Q1.y == Q2.y) && (Q2.y == Q3.y))
  { /* Degenerates to a point */
    Trans[0][0] = Trans[0][1] = Trans[1][0] = Trans[1][1] = 0;
    Trans[2][0] = Q1.x;
    Trans[2][1] = Q1.y;
    Trans[2][2] = 1;
    return(TRUE);
  }
  if ((Q1.x == Q2.x) && (Q1.y == Q2.y))
  { /* Same point - swap */
    temp = Q3.x; Q3.x = Q2.x; Q2.x = temp;
    temp = Q3.y; Q3.y = Q2.y; Q2.y = temp;
    temp = P3.x; P3.x = P2.x; P2.x = temp;
    temp = P3.y; P3.y = P2.y; P2.y = temp;
  }

  /* transform P1, Q1 to origin */
  P2P.x = P2.x - P1.x;
  P2P.y = P2.y - P1.y;
  P3P.x = P3.x - P1.x;
  P3P.y = P3.y - P1.y;
  Q2P.x = Q2.x - Q1.x;
  Q2P.y = Q2.y - Q1.y;
  Q3P.x = Q3.x - Q1.x;
  Q3P.y = Q3.y - Q1.y;

  /* Compute partial factors for matrix */
  LP2Psqrd = P2P.x * P2P.x + P2P.y * P2P.y;
  LQ2Psqrd = Q2P.x * Q2P.x + Q2P.y * Q2P.y;
  if((LP2Psqrd == 0.0) || (LQ2Psqrd == 0.0))
    return(FALSE); /* Degenerate triangle - two vertices identical */
  A = P3P.y * P2P.x - P3P.x * P2P.y;
  if(A == 0.0) /* Degenerate triangle - vertices are collinear */
    return(FALSE);
  B = (Q3P.x * Q2P.x + Q3P.y * Q2P.y) / LQ2Psqrd -
      (P3P.x * P2P.x + P3P.y * P2P.y) / LP2Psqrd;
  C = (Q2P.x * Q3P.y - Q3P.x * Q2P.y) / LQ2Psqrd;

  /* Load matrix */
  Trans[0][0] = (P2P.x * Q2P.x)/LP2Psqrd -
                         (P2P.y * Q2P.x * B - Q2P.y * P2P.y * C)/A;
  Trans[0][1] = (P2P.x * Q2P.y)/LP2Psqrd -
                         (P2P.y * Q2P.y * B + Q2P.x * P2P.y * C)/A;
  Trans[0][2] = 0;
  Trans[1][0] = (P2P.y * Q2P.x)/LP2Psqrd +
                         (P2P.x * Q2P.x * B - Q2P.y * P2P.x * C)/A;
  Trans[1][1] = (P2P.y * Q2P.y)/LP2Psqrd +
                         (P2P.x * Q2P.y * B + Q2P.x * P2P.x * C)/A;
  Trans[1][2] = 0;
  Trans[2][0] = Q1.x;
  Trans[2][1] = Q1.y;
  Trans[2][2] = 1;

  /* factor in the initial translation of P1 */
  Trans[2][0] = (-P1.x) * Trans[0][0] + (-P1.y) * Trans[1][0] + Trans[2][0];
  Trans[2][1] = (-P1.x) * Trans[0][1] + (-P1.y) * Trans[1][1] + Trans[2][1];
  Trans[2][2] = (-P1.x) * Trans[0][2] + (-P1.y) * Trans[1][2] + Trans[2][2];

/* Debug testing */
  tempx = P1.x * Trans[0][0] + P1.y * Trans[1][0] + Trans[2][0];
  tempy = P1.x * Trans[0][1] + P1.y * Trans[1][1] + Trans[2][1];
  if (!FEQUAL(tempx, Q1.x) || !FEQUAL(tempy, Q1.y))
    GSelog(0, "Error in triangle_transform: %lf, %lf to %lf %lf, not %lf %lf\n",
	   P1.x, P1.y, tempx, tempy, Q1.x, Q1.y);
  tempx = P2.x * Trans[0][0] + P2.y * Trans[1][0] + Trans[2][0];
  tempy = P2.x * Trans[0][1] + P2.y * Trans[1][1] + Trans[2][1];
  if (!FEQUAL(tempx, Q2.x) || !FEQUAL(tempy, Q2.y))
    GSelog(0, "Error in triangle_transform: %lf, %lf to %lf %lf, not %lf %lf\n",
	   P2.x, P2.y, tempx, tempy, Q2.x, Q2.y);
  tempx = P3.x * Trans[0][0] + P3.y * Trans[1][0] + Trans[2][0];
  tempy = P3.x * Trans[0][1] + P3.y * Trans[1][1] + Trans[2][1];
  if (!FEQUAL(tempx, Q3.x) || !FEQUAL(tempy, Q3.y))
    GSelog(0, "Error in triangle_transform: %lf, %lf to %lf %lf, not %lf %lf\n",
	   P3.x, P3.y, tempx, tempy, Q3.x, Q3.y);

  return(TRUE);
}
