/*================================ triang.c ==================================

Example program...
  
    Point2      p[100];
    Triangle    t[200];
    Edge        e[300];
    long        np, nt, ne;
  
    <Generate or read np points into array p.>
    <Generate or read ne edges into array e.>
    nt = triangulate(p, np, e, &ne, t);
    <You now have nt triangles in array t, and ne edges in array e.>


Things to do:
        Add more documentation; change variable names.
        Put triangle neighbors in order.
        Find a faster edge acceptance test...(yeah, right).

----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include "triang.h"

/* Uncomment the following line if you wish to have edges added in
** ascending order by length.
*/
/*#define SORT_EDGES*/

#undef  NULL
#define NULL          ( (void *) 0 )
#define EQUAL(a,b)    ( (a)->x == (b)->x  &&  (a)->y == (b)->y )

typedef struct {
    double      sql;
    Point2     *p, *q;
    }
    length;

static Point2      *Cen;     /* central point for function cmp_directions */
static Edge        *Ae;      /* the array of edges */
static Triangle    *At;      /* the array of triangles */
static long         Ne, Nt;  /* sizes of the above arrays */

/* status codes returned by function try_edge */
enum { ACCEPT, DUPLICATE, ZERO_LENGTH, CROSS };

/*---------------------------- Private Functions ---------------------------*/

static double
    qatan2(double y, double x)
    {
    /* A quick "arctangent" function.  Similar to atan2 in the standard
    ** C math library, it instead returns a piecewise-continuous linear
    ** function ranging from -4 to 4, not the true angle ranging from
    ** -pi to pi.  Used by cmp_directions.
    */

    if (x == 0.  &&  y == 0.)    return 0.;
    if (ABS(x) < ABS(y))    return (y<0.? -2. : 2.) - x/y;
    if (x > 0.)    return y/x;
    return (y<0.? -4. : 4.) + y/x;
    }



static int
    cmp_directions(const void *va, const void *vb)
    {
    /* Compare two point pointers based on the direction angles their
    ** points make with the central point *Cen.  Intended for use
    ** with the standard C function qsort.
    */

    Point2      *a, *b;
    double       dira, dirb;

    a = *((Point2 **) va);    b = *((Point2 **) vb);

    dira = qatan2(a->y - Cen->y, a->x - Cen->x);
    dirb = qatan2(b->y - Cen->y, b->x - Cen->x);
    return dira < dirb? -1 : (dira > dirb);
    }



static int
    cmp_lengths(const void *va, const void *vb)
    {
    /* Compare two length structures by their "sql" fields.
    ** Intended for use with the standard C function qsort.
    */

    double    sqla, sqlb;

    sqla = ((length *) va)->sql;
    sqlb = ((length *) vb)->sql;
    return sqla < sqlb? -1 : (sqla > sqlb);
    }



static int
    cmp_edges(const void *va, const void *vb)
    {
    /* Compare two edges based on the "indices" of their endpoints,
    ** primarily by p, secondarily by q.  Intended for use with
    ** the standard C functions bsearch and qsort.
    */

    Edge    *a, *b;

    a = (Edge *) va;    b = (Edge *) vb;
    
    if (a->p < b->p)         return -1;
    else if (a->p > b->p)    return  1;
    else return a->q < b->q? -1 : (a->q > b->q);
    }



static Edge *
    find_edge(Point2 *p, Point2 *q)
    {
    /* Find an edge, having endpoints p and q, in the sorted array
    ** of edges Ae[Ne].  Returns a pointer to the edge, or NULL if
    ** there is no such edge.
    */

    Edge    e;

    /* Edges in the array are always stored with p < q.  We must make
    ** sure the caller's edge has the same property.
    */
    if (p < q)    { e.p = p;  e.q = q; }
    else          { e.p = q;  e.q = p; }

    return ((Edge *)bsearch(&e, Ae, Ne, sizeof(e), cmp_edges));
    }



static int
    side_point(Point2 *r, Edge *e)
    {
    /* Return -1, 0, or 1, depending on whether r is below, on, or
    ** above the infinite line in which edge e resides.  "Above" and
    ** "below" are defined in this context by regarding p as "left"
    ** and q as "right".
    */

    double    cross; /* magnitude of cross product of relative vectors */

    cross = (e->q->x - e->p->x)*(r->y - e->p->y) -
            (e->q->y - e->p->y)*(r->x - e->p->x);

    return (cross < 0.)? -1 : (cross > 0.);
    }



static int
    try_edge(Point2 *p, Point2 *q)
    {
    /* Try to add an edge with endpoints p and q to the array Ae[Ne].
    ** One of the following result codes is returned:
    **     ACCEPT      => the edge has been accepted
    **     ZERO_LENGTH => p and q are the same point
    **     DUPLICATE   => the edge duplicates an accepted edge
    **     CROSS       => the edge crosses an accepted edge
    ** The edge is added to the array only on an ACCEPT result.
    */

    Edge     *e,             /* indexes through edge array */
             *f = Ae + Ne;   /* the tentative new edge */
    int       n;

    /* Ensure that the endpoints are not equal. */
    if (EQUAL(p, q))    return ZERO_LENGTH;

    /* Edges are always stored with p < q. */
    if (p < q)    { f->p = p;  f->q = q; }
    else          { f->p = q;  f->q = p; }

    /* Loop over all previously accepted edges. */
    for (e = Ae;  e < f;  e++)
        {
	/* n = number of endpoints common to edges e and f. */
	n = EQUAL(e->p,p) + EQUAL(e->p,q) + EQUAL(e->q,p) + EQUAL(e->q,q);
	if (n >= 2)    return DUPLICATE;

	/* If any of these tests succeed, edges e and f do not cross. */
	if (n == 1  ||
            MAX(e->p->x, e->q->x) < MIN(f->p->x, f->q->x)  ||
            MAX(f->p->x, f->q->x) < MIN(e->p->x, e->q->x)  ||
            MAX(e->p->y, e->q->y) < MIN(f->p->y, f->q->y)  ||
            MAX(f->p->y, f->q->y) < MIN(e->p->y, e->q->y)  ||
            side_point(e->p, f)*side_point(e->q, f) >= 0  ||
            side_point(f->p, e)*side_point(f->q, e) >= 0
            )
            continue;

	/* All the above tests failed, thus we found an edge crossing. */
	return CROSS;
	}

    /* If we make it to here, the edge is acceptable.  Thus, initialize
    ** the remaining fields in the edge structure, increment the total
    ** number of edges, and return ACCEPT.
    */
    f->t = f->u = NULL;    Ne++;    return ACCEPT;
    }



static void
    try_triangle(Point2 *p, Point2 *q, Point2 *r)
    {
    /* Try to add the triangle defined by points p, q, and r to the
    ** array At[Nt].  The triangle exists if there is an edge connecting
    ** q and r.  We know already that there are edges from p to q, and
    ** from p to r, because triangulate calls this function on adjacent
    ** neighbors of the central point p.
    */

    Triangle    *t = At + Nt;    /* the tentative new triangle */
    Edge        *e;

    /* Try to find the edge from q to r.  If it is not there, then there
    ** is no triangle to be added.
    */
    if (!(e = find_edge(q, r)))    return;

    /* Initialize fields of the new triangle. */
    t->nt = 0;
    t->t[0] = t->t[1] = t->t[2] = NULL;
    t->p[0] = p;    t->p[1] = q;    t->p[2] = r;

    /* Record this new triangle in each of the triangle's three edges.
    ** Increment the number of triangles.
    */
    if (e->t)  e->u = t;  else e->t = t;
    e = find_edge(p, q);    if (e->t)  e->u = t;  else e->t = t;
    e = find_edge(p, r);    if (e->t)  e->u = t;  else e->t = t;
    Nt++;
    }

/*---------------------------- Public Functions ----------------------------*/

long
    triangulate(Point2 *ap, long np, Edge *ae, long *ne, Triangle *at)
    {
    /* Triangulate the set of points ap[np] starting with an
    ** initial set of edges ae[*ne], adding additional edges
    ** as necessary and putting the triangles in triangle array at[].
    ** Returns the number of triangles generated, and puts the final
    ** number of edges in *ne.
    **
    ** The caller must provide space for the edges and triangles.
    ** The array ae[] should be of size 3*np, and the array at[] should
    ** be of size 2*np.
    **
    ** If the returned result is negative, there was an error:
    **     -1: invalid arguments
    **     -2: insufficient memory for temporary data
    **     -3: initial edges intersect; bad edge is returned in *ne
    */

    Point2    **app;           /* temporary array of point pointers */
    length     *len = NULL;    /* temporary array of edge lengths */
    long        i, j, n;

    /* Check passed arguments for problems. */
    if (!ap || np < 3 || !ae || !ne || *ne < 0 || !at)    return -1;

    /* Attempt to allocate temporary storage. */
    app = (void *) malloc(np * sizeof(app[0]));
    if (!app)    return -2;
#ifdef SORT_EDGES
    len = (void *) malloc( ((np*(np-1))/2) * sizeof(len[0]) );
#endif

    /* Initialize the global array variables, so that the private functions
    ** can access the edge and triangle arrays.  We start with an "empty"
    ** edge array, even though the caller has edges stored there.
    */
    Ae = ae;    At = at;    Ne = Nt = 0;

    /* Attempt to add each of the caller's edges.  A "cross" error in this
    ** stage is fatal.
    */
    for (i = 0;  i < *ne;  i++)
        if (try_edge(ae[i].p, ae[i].q) == CROSS) { *ne = i;  return -3; }

    /* Add each of the possible edges between every pair of points.
    ** We can ignore errors in this stage.  After trying all these edges,
    ** we will have our final set of Ne edges in array Ae.
    */
    if (len)
        {
	for (j = 1;  j < np;  j++)    for (i = 0;  i < j;  i++, len++)
	    {
	    double    dx, dy;

	    len->p = ap + i;    dx = ap[i].x - ap[j].x;
	    len->q = ap + j;    dy = ap[i].y - ap[j].y;
	    len->sql = dx*dx + dy*dy;
	    }

	n = (np*(np-1))/2;    len -= n;
	qsort(len, n, sizeof(len[0]), cmp_lengths);
	for (i = 0;  i < n;  i++)    try_edge(len[i].p, len[i].q);
	free(len);
	}
    else  for (i = 1;  i < np;  i++)    for (j = i;  j < np;  j++)
	try_edge(ap+j, ap+j-i);

    /* All edges have now been added.  Sort them by their endpoint indices.
    */
    qsort(Ae, Ne, sizeof(Ae[0]), cmp_edges);

    /* Collect neighbors of common central points into the buffer app.
    ** Sort these neighbors in anti-clockwise order around the central
    ** point.  Then attempt to add triangles composed of the central
    ** point and adjacent pairs of neighbors.
    */
    for (i = 0;  i < Ne;)
        {
	Cen = Ae[i].p;
        for (j = i;  j < Ne  &&  Ae[j].p == Cen;  j++)
            app[j-i] = Ae[j].q;
	qsort(app, n = j-i, sizeof(app[0]), cmp_directions);

	if (n == 2)    try_triangle(Cen, app[0], app[1]);
	else if (n > 2)
	    {
	    try_triangle(Cen, app[n-1], app[0]);
	    for (j = 1;  j < n;  j++)
                try_triangle(Cen, app[j-1], app[j]);
	    }

	i += n;
	}

    /* Link the triangles together, based on the information in the
    ** edge data structures.
    */
    for (i = 0;  i < Ne;  i++)
        {
	Triangle    *t = Ae[i].t, *u = Ae[i].u;

	if (t && u) { t->t[t->nt++] = u;  u->t[u->nt++] = t; }
	}

    /* The end... */
    free(app);    *ne = Ne;    return Nt;
    }

/*==========================================================================*/

